"""Added webhooks.

Revision ID: acb40dabc2c0
Revises: d40eba8ebc4c
Create Date: 2025-01-30 14:32:16.596842

"""
from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = "acb40dabc2c0"
down_revision: str | None = "d40eba8ebc4c"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade database to this revision from previous."""
    # fmt: off
    op.create_table('webhooks',
    sa.Column('id', sa.UUID(), nullable=False, comment='Webhook ID'),
    sa.Column('user_id', sa.String(), nullable=True, comment='The user ID associated with the webhook'),
    sa.Column('url', sa.String(), nullable=False, comment='The URL to which the webhook will send requests'),
    sa.Column('events', sa.JSON().with_variant(postgresql.JSONB(astext_type=sa.Text()), 'postgresql'), nullable=False, comment='List of events the webhook is subscribed to. If None, the webhook is subscribed to all events.'),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False, comment='Timestamp when row is inserted'),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False, comment='Timestamp when row is updated'),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_webhooks'))
    )
    with op.batch_alter_table('webhooks', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_webhooks_user_id'), ['user_id'], unique=False)
    # fmt: on


def downgrade() -> None:
    """Downgrade database from this revision to previous."""
    with op.batch_alter_table("webhooks", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_webhooks_user_id"))

    op.drop_table("webhooks")



from pathlib import Path

from pydantic import BaseModel, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from aana.configs.db import DbSettings, SQLiteConfig
from aana.core.models.base import pydantic_protected_fields
from aana.storage.op import DbType


class TestSettings(BaseModel):
    """A pydantic model for test settings.

    Attributes:
        test_mode (bool): Flag indicating if the SDK is in test mode.
        save_expected_output (bool): Flag indicating if the expected output should be saved (to create test cases).
    """

    test_mode: bool = False
    save_expected_output: bool = False


class TaskQueueSettings(BaseModel):
    """A pydantic model for task queue settings.

    Attributes:
        enabled (bool): Flag indicating if the task queue is enabled.
        num_workers (int): The number of workers in the task queue.
        execution_timeout (int): The maximum execution time for a task in seconds.
            After this time, if the task is still running,
            it will be considered as stuck and will be reassign to another worker.
        heartbeat_timeout (int): The maximum time between heartbeats in seconds.
        max_retries (int): The maximum number of retries for a task.
        maximum_active_tasks_per_user (int): The maximum number of active tasks per user (only applicable in the API service).
    """

    enabled: bool = True
    num_workers: int = 4
    execution_timeout: int = 600
    heartbeat_timeout: int = 60
    max_retries: int = 3
    maximum_active_tasks_per_user: int = 25


class ApiServiceSettings(BaseModel):
    """A pydantic model for API service settings.

    Attributes:
        enabled (bool): Flag indicating if the API service is enabled.
        lago_url (str): The URL of the LAGO API service.
        lago_api_key (str): The API key for the LAGO API service.
    """

    enabled: bool = False
    lago_url: str | None = None
    lago_api_key: str | None = None


class WebhookSettings(BaseModel):
    """A pydantic model for webhook settings.

    Attributes:
        retry_attempts (int): The number of retry attempts for webhook delivery.
        hmac_secret (str): The secret key for HMAC signature generation.
    """

    retry_attempts: int = 5
    hmac_secret: str = "webhook_secret" # noqa: S105


class Settings(BaseSettings):
    """A pydantic model for SDK settings.

    Attributes:
        tmp_data_dir (Path): The temporary data directory.
        image_dir (Path): The temporary image directory.
        video_dir (Path): The temporary video directory.
        audio_dir (Path): The temporary audio directory.
        model_dir (Path): The temporary model directory.
        num_workers (int): The number of web workers.
        openai_endpoint_enabled (bool): Flag indicating if the OpenAI-compatible endpoint is enabled. Enabled by default.
        include_stacktrace (bool): Flag indicating if stacktrace should be included in error messages. Enabled by default.
        task_queue (TaskQueueSettings): The task queue settings.
        db_config (DbSettings): The database configuration.
        test (TestSettings): The test settings.
    """

    tmp_data_dir: Path = Path("/tmp/aana_data")  # noqa: S108
    image_dir: Path | None = None
    video_dir: Path | None = None
    audio_dir: Path | None = None
    model_dir: Path | None = None

    num_workers: int = 2

    openai_endpoint_enabled: bool = True
    include_stacktrace: bool = True

    task_queue: TaskQueueSettings = TaskQueueSettings()

    db_config: DbSettings = DbSettings()

    test: TestSettings = TestSettings()

    api_service: ApiServiceSettings = ApiServiceSettings()

    api_service_db_config: DbSettings = DbSettings(
        datastore_type=DbType.SQLITE,
        datastore_config=SQLiteConfig(path="/var/lib/aana_api_service_data"),
    )

    webhook: WebhookSettings = WebhookSettings()

    @model_validator(mode="after")
    def setup_resource_directories(self):
        """Create the resource directories if they do not exist."""
        if self.image_dir is None:
            self.image_dir = self.tmp_data_dir / "images"
        if self.video_dir is None:
            self.video_dir = self.tmp_data_dir / "videos"
        if self.audio_dir is None:
            self.audio_dir = self.tmp_data_dir / "audios"
        if self.model_dir is None:
            self.model_dir = self.tmp_data_dir / "models"

        self.tmp_data_dir.mkdir(parents=True, exist_ok=True)
        self.image_dir.mkdir(parents=True, exist_ok=True)
        self.video_dir.mkdir(parents=True, exist_ok=True)
        self.audio_dir.mkdir(parents=True, exist_ok=True)
        self.model_dir.mkdir(parents=True, exist_ok=True)
        return self

    model_config = SettingsConfigDict(
        protected_namespaces=("settings", *pydantic_protected_fields),
        env_nested_delimiter="__",
        env_ignore_empty=True,
        env_file=".env",
        extra="ignore",
    )


settings = Settings()



from typing import Any

from pydantic import BaseModel, Field

from aana.core.models.api_service import ApiKey
from aana.storage.models.task import Status as TaskStatus
from aana.storage.models.task import TaskEntity


class TaskInfo(BaseModel):
    """Task information.

    Attributes:
        id (str): The task ID.
        endpoint (str): The endpoint to which the task is assigned.
        data (Any): The task data.
        status (TaskStatus): The task status.
        result (Any): The task result.
    """

    id: str = Field(..., description="The task ID.")
    endpoint: str = Field(
        ..., description="The endpoint to which the task is assigned."
    )
    data: Any = Field(..., description="The task data.")
    status: TaskStatus = Field(..., description="The task status.")
    result: Any = Field(None, description="The task result.")

    @classmethod
    def from_entity(cls, task: TaskEntity) -> "TaskInfo":
        """Create a TaskInfo from a TaskEntity."""
        # Prepare data (remove ApiKey, None values, etc.)
        task_data = {}
        for key, value in task.data.items():
            if value is None or isinstance(value, ApiKey):
                continue
            task_data[key] = value

        return TaskInfo(
            id=str(task.id),
            endpoint=task.endpoint,
            data=task_data,
            status=task.status,
            result=task.result,
        )



# ruff: noqa: S101, S113
import hashlib
import hmac
import json
from typing import Annotated, TypedDict

import requests
from pydantic import Field

from aana.api.api_generation import Endpoint
from aana.configs.settings import settings as aana_settings

TextList = Annotated[list[str], Field(description="List of text to lowercase.")]


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: list[str]


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def run(self, text: TextList) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        return {"text": [t.lower() for t in text]}


deployments = []

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    }
]


def test_webhooks(create_app, httpserver):
    """Test webhooks."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}/api/ready")
    assert response.status_code == 200, response.text
    assert response.json() == {"ready": True}

    # Setup the webhook listener
    def webhook_listener(request):
        payload = request.json

        # Validate the HMAC signature
        secret_key = aana_settings.webhook.hmac_secret
        signature = request.headers.get("X-Signature")
        payload_str = json.dumps(payload, separators=(",", ":"))
        actual_signature = hmac.new(
            secret_key.encode(), payload_str.encode(), hashlib.sha256
        ).hexdigest()
        assert signature == actual_signature
        assert payload["event"] == "task.completed"

    httpserver.expect_request("/webhooks").respond_with_handler(webhook_listener)

    # Register the webhook
    data = {
        "url": f"http://localhost:{httpserver.port}/webhooks",
        "events": ["task.completed"],
    }
    response = requests.post(
        f"http://localhost:{port}/webhooks",
        json=data,
    )
    assert response.status_code == 201, response.text

    # Test lowercase endpoint
    data = {"text": ["Hello World!", "This is a test."]}
    response = requests.post(
        f"http://localhost:{port}/lowercase",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200, response.text
    lowercase_text = response.json().get("text")
    assert lowercase_text == ["hello world!", "this is a test."]

    # Defer endpoint execution
    response = requests.post(
        f"http://localhost:{port}/lowercase?defer=True",
        data={"body": json.dumps(data)},
    )
    assert response.status_code == 200, response.text

    httpserver.check_assertions()
    httpserver.check_handler_errors()


def test_webhook_crud(create_app):
    """Test webhook CRUD operations."""
    aana_app = create_app(deployments, endpoints)
    port = aana_app.port
    base_url = f"http://localhost:{port}"

    # Clear existing webhooks
    response = requests.get(f"{base_url}/webhooks")
    assert response.status_code == 200
    webhooks = response.json()["webhooks"]
    for webhook in webhooks:
        response = requests.delete(f"{base_url}/webhooks/{webhook['id']}")
        assert response.status_code == 200

    # Test Create
    webhook_data = {
        "url": "http://example.com/webhook",
        "events": ["task.completed", "task.failed"],
    }
    response = requests.post(f"{base_url}/webhooks", json=webhook_data)
    assert response.status_code == 201
    webhook_id = response.json()["id"]
    assert response.json()["url"] == webhook_data["url"]
    assert response.json()["events"] == webhook_data["events"]

    # Test Read (List)
    response = requests.get(f"{base_url}/webhooks")
    assert response.status_code == 200
    webhooks = response.json()["webhooks"]
    assert len(webhooks) == 1
    assert webhooks[0]["id"] == webhook_id

    # Test Read (Single)
    response = requests.get(f"{base_url}/webhooks/{webhook_id}")
    assert response.status_code == 200
    assert response.json()["id"] == webhook_id
    assert response.json()["url"] == webhook_data["url"]

    # Test Update
    update_data = {"url": "http://example.com/webhook2", "events": ["task.completed"]}
    response = requests.put(f"{base_url}/webhooks/{webhook_id}", json=update_data)
    assert response.status_code == 200
    assert response.json()["url"] == update_data["url"]
    assert response.json()["events"] == update_data["events"]

    # Test Delete
    response = requests.delete(f"{base_url}/webhooks/{webhook_id}")
    assert response.status_code == 200

    # Verify webhook is deleted
    response = requests.get(f"{base_url}/webhooks/{webhook_id}")
    assert response.status_code == 404

    # Test validation for URL
    response = requests.post(f"{base_url}/webhooks", json={"url": "invalid-url"})
    assert response.status_code == 422
    assert response.json()["error"] == "ValidationError"



from typing import Annotated

from fastapi import Depends, Request

from aana.configs.settings import settings as aana_settings
from aana.core.models.api_service import ApiKey
from aana.exceptions.api_service import AdminOnlyAccess, InactiveSubscription


def is_admin(request: Request) -> bool:
    """Check if the user is an admin.

    Args:
        request (Request): The request object

    Returns:
        bool: True if the user is an admin, False otherwise
    """
    if aana_settings.api_service.enabled:
        api_key_info: ApiKey = request.state.api_key_info
        return api_key_info.is_admin if api_key_info else False
    return True


def require_admin_access(request: Request) -> bool:
    """Check if the user is an admin. If not, raise an exception.

    Args:
        request (Request): The request object

    Raises:
        AdminOnlyAccess: If the user is not an admin
    """
    _is_admin = is_admin(request)
    if not _is_admin:
        raise AdminOnlyAccess()
    return True


def extract_api_key_info(request: Request) -> ApiKey | None:
    """Get the API key info dependency."""
    return getattr(request.state, "api_key_info", None)


def extract_user_id(request: Request) -> str | None:
    """Get the user ID dependency."""
    api_key_info = extract_api_key_info(request)
    return api_key_info.user_id if api_key_info else None


def require_active_subscription(request: Request) -> bool:
    """Check if the user has an active subscription. If not, raise an exception.

    Args:
        request (Request): The request object

    Raises:
        InactiveSubscription: If the user does not have an active subscription
    """
    if aana_settings.api_service.enabled:
        api_key_info: ApiKey = request.state.api_key_info
        if not api_key_info.is_subscription_active:
            raise InactiveSubscription(key=api_key_info.api_key)
    return True


AdminAccessDependency = Annotated[bool, Depends(require_admin_access)]
""" Dependency to check if the user is an admin. If not, it will raise an exception. """

IsAdminDependency = Annotated[bool, Depends(is_admin)]
""" Dependency to check if the user is an admin. """

UserIdDependency = Annotated[str | None, Depends(extract_user_id)]
""" Dependency to get the user ID. """

ApiKeyInfoDependency = Annotated[ApiKey | None, Depends(extract_api_key_info)]
""" Dependency to get the API key info. """

ActiveSubscriptionRequiredDependency = Annotated[
    bool, Depends(require_active_subscription)
]
""" Dependency to check if the user has an active subscription. If not, it will raise an exception. """



# ruff: noqa: S101, S108
from pathlib import Path

from aana.configs.settings import Settings


def test_default_tmp_data_dir():
    """Test that the default temporary data directory is set correctly."""
    settings = Settings()
    assert settings.tmp_data_dir == Path("/tmp/aana_data")


def test_custom_tmp_data_dir(monkeypatch):
    """Test that the custom temporary data directory with environment variable is set correctly."""
    test_path = "/tmp/override/path"
    monkeypatch.setenv("TMP_DATA_DIR", test_path)
    settings = Settings()
    assert settings.tmp_data_dir == Path(test_path)


def test_changing_tmp_data_dir():
    """Test that changing the temporary data directory is reflected in the other directories."""
    new_tmp_data_dir = Path("/tmp/new_tmp_data_dir")
    settings = Settings(tmp_data_dir=new_tmp_data_dir)

    assert settings.tmp_data_dir == new_tmp_data_dir
    assert settings.image_dir == new_tmp_data_dir / "images"
    assert settings.video_dir == new_tmp_data_dir / "videos"
    assert settings.audio_dir == new_tmp_data_dir / "audios"
    assert settings.model_dir == new_tmp_data_dir / "models"

    # Check that we can change the image directory independently
    new_image_dir = Path("/tmp/new_image_dir")
    settings = Settings(tmp_data_dir=new_tmp_data_dir, image_dir=new_image_dir)
    assert settings.tmp_data_dir == new_tmp_data_dir
    assert settings.image_dir == new_image_dir

    # Check that we can change the video directory independently
    new_video_dir = Path("/tmp/new_video_dir")
    settings = Settings(tmp_data_dir=new_tmp_data_dir, video_dir=new_video_dir)
    assert settings.tmp_data_dir == new_tmp_data_dir
    assert settings.video_dir == new_video_dir

    # Check that we can change the audio directory independently
    new_audio_dir = Path("/tmp/new_audio_dir")
    settings = Settings(tmp_data_dir=new_tmp_data_dir, audio_dir=new_audio_dir)
    assert settings.tmp_data_dir == new_tmp_data_dir
    assert settings.audio_dir == new_audio_dir

    # Check that we can change the model directory independently
    new_model_dir = Path("/tmp/new_model_dir")
    settings = Settings(tmp_data_dir=new_tmp_data_dir, model_dir=new_model_dir)
    assert settings.tmp_data_dir == new_tmp_data_dir
    assert settings.model_dir == new_model_dir



from dataclasses import dataclass
from typing import Any, TypedDict

import torch
from huggingface_hub.utils import GatedRepoError
from pydantic import BaseModel, ConfigDict, Field
from ray import serve

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_protected_fields
from aana.core.models.time import TimeInterval
from aana.core.models.vad import VadParams, VadSegment
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.utils.lazy_import import LazyImport

with LazyImport(
    "Run 'pip install pyannote-audio' or 'pip install aana[asr]'"
) as pyannote_imports:
    from pyannote.audio import Model

    from aana.processors.vad import BinarizeVadScores, VoiceActivitySegmentation


@dataclass
class SegmentX:
    """The vad segment format with optional speaker.

    Attributes:
        start (float): The start time of the segment.
        end (float): The end time of the segment.
        speaker (str, optional): The speaker of the segment. Defaults to "".
    """

    start: float
    end: float
    speaker: str = ""


class VadOutput(TypedDict):
    """The output of the VAD model.

    Attributes:
        segments (list[VadSegment]): The VAD segments.
    """

    segments: list[VadSegment]


class VadConfig(BaseModel):
    """The configuration for the vad deployment.

    Attributes:
        model_id (str): The model id.
        onset (float): The onset threshold. Defaults to 0.500.
        offset (float): The offset threshold. Defaults to 0.363.
        min_duration_on (float): The minimum duration for voice activity. Defaults to 0.1.
        min_duration_off (float): The minimum duration for silence. Defaults to 0.1.
        sample_rate (int): The sample rate of the audio. Defaults to 16000.
    """

    model_id: str = Field(
        default="pyannote/segmentation",
        description="The VAD model id.",
    )

    # default_parameters for the model on initialization:
    onset: float = Field(
        default=0.500,
        ge=0.0,
        description="Threshold to decide a positive voice activity.",
    )

    offset: float = Field(
        default=0.363,
        ge=0.0,
        description="Threshold to consider as a silence region.",
    )

    min_duration_on: float = Field(
        default=0.1, ge=0.0, description="Minimum voiced duration."
    )

    min_duration_off: float = Field(
        default=0.1, ge=0.0, description="Minimum duration to consider as silence."
    )

    sample_rate: int = Field(default=16000, description="Sample rate of the audio.")

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment(health_check_timeout_s=180)
class VadDeployment(BaseDeployment):
    """Deployment to serve VAD models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and instantiate vad_pipeline.

        The configuration should conform to the VadConfig schema.

        """
        pyannote_imports.check()
        config_obj = VadConfig(**config)
        self.hyperparameters = {
            "onset": config_obj.onset,
            "offset": config_obj.offset,
            "min_duration_on": config_obj.min_duration_on,
            "min_duration_off": config_obj.min_duration_off,
        }
        self.sample_rate = config_obj.sample_rate
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # for consistency across multiple runs
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(42)

        self.model_id = config_obj.model_id
        try:
            # load model using pyannote Pipeline
            self.segmentation_model = Model.from_pretrained(
                self.model_id, use_auth_token=None
            )
            self.vad_pipeline = VoiceActivitySegmentation(
                segmentation=self.segmentation_model, device=torch.device(self.device)
            )

        except Exception as e:
            raise GatedRepoError(
                message=f"This repository is private and requires a token to accept user conditions and access models in {self.model_id} pipeline."
            ) from e
        self.vad_pipeline.instantiate(self.hyperparameters)

    async def __merge(
        self, segments: list[dict], params: VadParams
    ) -> list[VadSegment]:
        """Merge operation to combine vad segments into chunk_size segments.

        Args:
            segments (list[dict]): The list of segments to merge/split.
            params (VadParams): The parameters for the vad model.  #dict[str, Any]

        Returns:
            list[VadSegment]: The list of segments.
        """
        curr_end = 0.0
        merged_segments: list[VadSegment] = []
        seg_idxs: list[tuple[float, float]] = []
        speaker_idxs: list[str] = []

        if not params.chunk_size > 0:
            chunk_error = "Expected positive value."
            raise ValueError(f"{chunk_error}")

        binarize = BinarizeVadScores(
            max_duration=params.chunk_size,
            onset=self.hyperparameters["onset"],
            offset=self.hyperparameters["offset"],
        )
        segments = binarize(segments)

        segments_list = [
            SegmentX(
                max(0.0, speech_turn.start - 0.1), speech_turn.end + 0.1, "UNKNOWN"
            )
            for speech_turn in segments.itersegments()
        ]
        if len(segments_list) == 0:
            # No active speech found in audio.
            return []

        # To make sure the starting point is the start of the segment.
        curr_start = segments_list[0].start

        for seg in segments_list:
            if seg.end - curr_start > params.chunk_size and curr_end - curr_start > 0:
                merged_segments.append(
                    VadSegment(
                        time_interval=TimeInterval(start=curr_start, end=curr_end),
                        segments=seg_idxs,
                    )
                )
                curr_start = seg.start
                seg_idxs = []
                speaker_idxs = []
            curr_end = seg.end
            seg_idxs.append((seg.start, seg.end))
            speaker_idxs.append(seg.speaker)

        merged_segments.append(
            VadSegment(
                time_interval=TimeInterval(start=curr_start, end=curr_end),
                segments=seg_idxs,
            )
        )
        return merged_segments

    async def __inference(self, audio: Audio) -> list[dict]:
        """Perform voice activity detection inference on the Audio with the vad model.

        Args:
            audio (Audio): The audio to perform vad.

        Returns:
            vad_segments (list[dict]): The list of vad segments.

        Raises:
            InferenceException: If the vad inference fails.
        """
        audio_array = audio.get_numpy()
        vad_input = {
            "waveform": torch.from_numpy(audio_array).unsqueeze(0),
            "sample_rate": self.sample_rate,
        }

        try:
            vad_segments = self.vad_pipeline(vad_input)
        except Exception as e:
            raise InferenceException(self.model_id, str(e)) from e

        return vad_segments

    @exception_handler
    async def asr_preprocess_vad(
        self, audio: Audio, params: VadParams | None = None
    ) -> VadOutput:
        """Perform vad inference to get vad segments and further processing (split and merge) to get segments for batched asr inference.

        Args:
            audio (Audio): The audio to perform vad.
            params (VadParams): The parameters for the vad model.

        Returns:
            VadOutput: Output vad segments to the asr model.

        Raises:
            InferenceException: If the vad inference fails.
        """
        if not params:
            params = VadParams()

        intermediate_vad_segments = await self.__inference(audio)
        asr_vad_segments = await self.__merge(
            intermediate_vad_segments,
            params,
        )
        return VadOutput(segments=asr_vad_segments)



from uuid import UUID

from sqlalchemy import func, or_
from sqlalchemy.orm import Session

from aana.exceptions.runtime import InvalidWebhookEventType
from aana.storage.models.webhook import WebhookEntity, WebhookEventType
from aana.storage.repository.base import BaseRepository


class WebhookRepository(BaseRepository[WebhookEntity]):
    """Repository for webhooks."""

    def __init__(self, session: Session):
        """Constructor."""
        super().__init__(session, WebhookEntity)

    def read(self, item_id: str | UUID, check: bool = True) -> WebhookEntity | None:
        """Reads a single webhook from the database.

        Args:
            item_id (str | UUID): ID of the webhook to retrieve.
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            The corresponding entity from the database if found.

        Raises:
            NotFoundException if the entity is not found and `check` is True.
        """
        try:
            if isinstance(item_id, str):
                item_id = UUID(item_id)
        except ValueError:
            return None
        return super().read(item_id, check)

    def delete(self, item_id: str | UUID, check: bool = True) -> WebhookEntity | None:
        """Delete a webhook from the database.

        Args:
            item_id (str | UUID): The ID of the webhook to delete.
            check (bool): whether to raise if the entity is not found (defaults to True).

        Returns:
            WebhookEntity: The deleted webhook.

        Raises:
            NotFoundException: The id does not correspond to a record in the database.
        """
        try:
            if isinstance(item_id, str):
                item_id = UUID(item_id)
        except ValueError:
            return None
        return super().delete(item_id, check)

    def save(self, webhook: WebhookEntity) -> WebhookEntity:
        """Save a webhook to the database.

        Args:
            webhook (WebhookEntity): The webhook to save.

        Returns:
            WebhookEntity: The saved webhook.
        """
        if webhook.events is None:
            webhook.events = []

        # Check if events are in WebhookEventType enum
        if webhook.events:
            try:
                webhook.events = [WebhookEventType(event) for event in webhook.events]
            except ValueError as e:
                raise InvalidWebhookEventType(event_type=e.args[0]) from e

        self.session.add(webhook)
        self.session.commit()
        return webhook

    def get_webhooks(
        self, user_id: str | None, event_type: str | None = None
    ) -> list[WebhookEntity]:
        """Get webhooks for a user.

        Args:
            user_id (str | None): The user ID. If None, get system-wide webhooks.
            event_type (str | None): Filter webhooks by event type. If None, return all webhooks.

        Returns:
            List[WebhookEntity]: The list of webhooks.
        """
        query = self.session.query(WebhookEntity).filter_by(user_id=user_id)
        if event_type:
            if self.session.bind.dialect.name == "postgresql":
                query = query.filter(
                    or_(
                        WebhookEntity.events.op("@>")([event_type]),
                        WebhookEntity.events == [],
                    )
                )
            elif self.session.bind.dialect.name == "sqlite":
                events_func = func.json_each(WebhookEntity.events).table_valued(
                    "value", joins_implicitly=True
                )
                query = query.filter(
                    or_(
                        self.session.query(events_func)
                        .filter(events_func.c.value == event_type)
                        .exists(),
                        WebhookEntity.events == "[]",
                    )
                )
            else:
                raise NotImplementedError(
                    f"Filtering by event type is not supported for {self.session.bind.dialect.name}"
                )

        return query.all()



# ruff: noqa: S101, S113
import json
import os
from datetime import datetime, timedelta, timezone
from typing import TypedDict

import pytest
import requests
from sqlalchemy.orm import Session

from aana.api.api_generation import Endpoint
from aana.core.models.api_service import ApiKeyType
from aana.storage.models.api_key import ApiKeyEntity

ACTIVE_API_KEY = "1234567890"
INACTIVE_API_KEY = "0000000000"
NON_EXISTENT_API_KEY = "1111111111"


@pytest.fixture(scope="module")
def enable_api_service():
    """Enable the API service."""
    os.environ["API_SERVICE__ENABLED"] = "True"
    os.environ["API_SERVICE__LAGO_URL"] = "http://localhost:3000"
    os.environ["API_SERVICE__LAGO_API_KEY"] = "test"
    yield
    # Cleanup after test
    os.environ.pop("API_SERVICE__ENABLED", None)
    os.environ.pop("API_SERVICE__LAGO_URL", None)
    os.environ.pop("API_SERVICE__LAGO_API_KEY", None)


@pytest.fixture(scope="module")
def add_test_api_keys():
    """Add test API keys to the database."""
    from aana.configs.settings import settings

    engine = settings.api_service_db_config.get_engine()
    with Session(engine) as session:
        # fmt: off
        session.add_all(
            [
                ApiKeyEntity(user_id="1", is_subscription_active=True, api_key=ACTIVE_API_KEY, key_id=ACTIVE_API_KEY, subscription_id="sub1", expired_at=datetime.now(tz=timezone.utc) + timedelta(days=180)),
                ApiKeyEntity(user_id="2", is_subscription_active=False, api_key=INACTIVE_API_KEY, key_id=INACTIVE_API_KEY, subscription_id="sub2", expired_at=datetime.now(tz=timezone.utc) + timedelta(days=180)),
            ]
        )
        # fmt: on
        session.commit()


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: str
    api_key: ApiKeyType


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def run(self, text: str, api_key: ApiKeyType) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (str): The list of text to lowercase
            api_key (ApiKey): The API key info

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = text.lower()
        return {"text": lowercase_output, "api_key": api_key}


class UppercaseEndpointOutput(TypedDict):
    """The output of the uppercase endpoint."""

    text: str


class UppercaseEndpoint(Endpoint):
    """Uppercase endpoint (withouth the API key)."""

    async def run(self, text: str) -> UppercaseEndpointOutput:
        """Uppercase the text.

        Args:
            text (str): The list of text to uppercase

        Returns:
            UppercaseEndpointOutput: The uppercase texts
        """
        uppercase_output = text.upper()
        return {"text": uppercase_output}


deployments = []

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "active_subscription_required": True,
        "endpoint_cls": LowercaseEndpoint,
    },
    {
        "name": "uppercase",
        "path": "/uppercase",
        "summary": "Uppercase text",
        "endpoint_cls": UppercaseEndpoint,
    },
]


def test_app(enable_api_service, create_app, add_test_api_keys):
    """Test the Ray Serve app."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port
    route_prefix = ""

    headers = {
        "x-api-key": ACTIVE_API_KEY,
    }

    # Check that the server is ready
    response = requests.get(
        f"http://localhost:{port}{route_prefix}/api/ready", headers=headers
    )
    assert response.status_code == 200, response.text
    assert response.json() == {"ready": True}, response.text

    # Test lowercase endpoint (endpoint with API key)
    data = {"text": "Hello World! This is a test."}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
        headers=headers,
    )
    assert response.status_code == 200, response.text
    lowercase_text = response.json().get("text")
    assert lowercase_text == "hello world! this is a test."

    api_key_in_response = response.json().get("api_key")
    assert api_key_in_response["api_key"] == ACTIVE_API_KEY, api_key_in_response

    # Test uppercase endpoint (endpoint without API key)
    data = {"text": "Hello World! This is a test."}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/uppercase",
        data={"body": json.dumps(data)},
        headers=headers,
    )
    assert response.status_code == 200, response.text
    uppercase_text = response.json().get("text")
    assert uppercase_text == "HELLO WORLD! THIS IS A TEST."

    # Test with inactive API key
    headers = {
        "x-api-key": INACTIVE_API_KEY,
    }
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
        headers=headers,
    )
    assert response.status_code == 400, response.text
    assert response.json().get("error") == "InactiveSubscription", response.json()

    # Test with non-existent API key
    headers = {
        "x-api-key": NON_EXISTENT_API_KEY,
    }
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
        headers=headers,
    )
    assert response.status_code == 400, response.text
    assert response.json().get("error") == "ApiKeyNotFound", response.json()

    # Test with missing API key
    headers = {}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/lowercase",
        data={"body": json.dumps(data)},
        headers=headers,
    )
    assert response.status_code == 400, response.text
    assert response.json().get("error") == "ApiKeyNotProvided", response.json()



from typing import Annotated

from fastapi import Depends
from sqlalchemy.orm import Session, sessionmaker

from aana.configs.settings import settings
from aana.storage.models.api_key import ApiServiceBase
from aana.storage.models.base import BaseEntity

__all__ = ["get_db", "get_session"]

engine = settings.db_config.get_engine()

if settings.api_service.enabled:
    api_service_engine = settings.api_service_db_config.get_engine()
    SessionLocal = sessionmaker(
        autocommit=False,
        autoflush=False,
        binds={ApiServiceBase: api_service_engine, BaseEntity: engine},
        bind=engine,  # Default engine
    )

else:
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def get_session() -> Session:
    """Get a new SQLAlchemy Session object.

    Returns:
        Session: SQLAlchemy Session object.
    """
    return SessionLocal()


def get_db():
    """Get a database session."""
    db = get_session()
    try:
        yield db
    finally:
        db.close()


GetDbDependency = Annotated[Session, Depends(get_db)]
""" Dependency to get a database session. """



# ruff: noqa: S101

import pytest

from aana.exceptions.runtime import InvalidWebhookEventType
from aana.storage.models.webhook import WebhookEntity
from aana.storage.repository.webhook import WebhookRepository


@pytest.fixture
def webhook_entities():
    """Create webhook entities for testing."""
    # fmt: off
    webhooks = [
        WebhookEntity(user_id="user1", url="https://example1.com", events=["task.completed", "task.failed"]),
        WebhookEntity(user_id="user1", url="https://example2.com", events=["task.failed"]),
        WebhookEntity(user_id="user2", url="https://example3.com", events=["task.completed", "task.failed"]),
        WebhookEntity(user_id=None, url="https://example4.com", events=["task.failed"]),  # System webhook for task.failed
        WebhookEntity(user_id=None, url="https://example5.com", events=None),  # System webhook for all events
    ]
    return webhooks
    # fmt: on


def test_save_webhook(db_session, webhook_entities):
    """Test saving a webhook."""
    webhook_repo = WebhookRepository(db_session)
    for webhook in webhook_entities:
        saved_webhook = webhook_repo.save(webhook)
        assert saved_webhook.id is not None
        assert saved_webhook.user_id == webhook.user_id
        assert saved_webhook.url == webhook.url
        assert saved_webhook.events == webhook.events

        retrieved_webhook = webhook_repo.read(saved_webhook.id)
        assert retrieved_webhook.id == saved_webhook.id
        assert retrieved_webhook.user_id == saved_webhook.user_id
        assert retrieved_webhook.url == saved_webhook.url

    # Test saving a webhook with invalid event type
    with pytest.raises(InvalidWebhookEventType):
        webhook = WebhookEntity(
            user_id="user1", url="https://example7.com", events=["invalid.event"]
        )
        webhook_repo.save(webhook)


def test_get_webhooks(db_session, webhook_entities):
    """Test fetching webhooks."""
    webhook_repo = WebhookRepository(db_session)

    # Save webhooks
    for webhook in webhook_entities:
        webhook_repo.save(webhook)

    # Test webhooks with user ID set to None
    webhooks = webhook_repo.get_webhooks(user_id=None)
    assert len(webhooks) == 2
    assert {webhook.url for webhook in webhooks} == {
        "https://example4.com",
        "https://example5.com",
    }

    # Test webhooks with user ID set to "user1"
    webhooks = webhook_repo.get_webhooks(user_id="user1")
    assert len(webhooks) == 2
    assert {webhook.url for webhook in webhooks} == {
        "https://example1.com",
        "https://example2.com",
    }

    # Test webhooks with user ID set to "user2"
    webhooks = webhook_repo.get_webhooks(user_id="user2")
    assert len(webhooks) == 1

    # Test webhooks with user ID set to "user3"
    webhooks = webhook_repo.get_webhooks(user_id="user3")
    assert len(webhooks) == 0

    # Test webhooks with user ID set to None and event type set to "task.failed"
    webhooks = webhook_repo.get_webhooks(user_id=None, event_type="task.failed")
    assert len(webhooks) == 2
    assert {webhook.url for webhook in webhooks} == {
        "https://example4.com",
        "https://example5.com",
    }

    # Test webhooks with user ID set to "user1" and event type set to "task.failed"
    webhooks = webhook_repo.get_webhooks(user_id="user1", event_type="task.failed")
    assert len(webhooks) == 2
    assert {webhook.url for webhook in webhooks} == {
        "https://example1.com",
        "https://example2.com",
    }

    # Test webhooks with user ID set to "user1" and event type set to "task.completed"
    webhooks = webhook_repo.get_webhooks(user_id="user1", event_type="task.completed")
    assert len(webhooks) == 1
    assert {webhook.url for webhook in webhooks} == {"https://example1.com"}



import json
import time
from uuid import uuid4

import ray
from fastapi import APIRouter
from fastapi.responses import StreamingResponse

from aana.api.responses import AanaJSONResponse
from aana.configs.settings import settings as aana_settings
from aana.core.models.chat import ChatCompletion, ChatCompletionRequest, ChatDialog
from aana.core.models.sampling import SamplingParams
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle

router = APIRouter(
    tags=["openai-api"], include_in_schema=aana_settings.openai_endpoint_enabled
)


@router.post(
    "/chat/completions",
    response_model=ChatCompletion,
)
async def chat_completions(request: ChatCompletionRequest):
    """Handle chat completions requests for OpenAI compatible API."""
    if not aana_settings.openai_endpoint_enabled:
        return AanaJSONResponse(
            content={
                "error": {"message": "The OpenAI-compatible endpoint is not enabled."}
            },
            status_code=404,
        )

    async def _async_chat_completions(
        handle: AanaDeploymentHandle,
        dialog: ChatDialog,
        sampling_params: SamplingParams,
    ):
        async for response in handle.chat_stream(
            dialog=dialog, sampling_params=sampling_params
        ):
            chunk = {
                "id": f"chatcmpl-{uuid4().hex}",
                "object": "chat.completion.chunk",
                "model": request.model,
                "created": int(time.time()),
                "choices": [
                    {
                        "index": 0,
                        "delta": {"content": response["text"], "role": "assistant"},
                    }
                ],
            }
            yield f"data: {json.dumps(chunk)}\n\n"
        yield "data: [DONE]\n\n"

    # Check if the deployment exists
    try:
        handle = await AanaDeploymentHandle.create(request.model)
    except ray.serve.exceptions.RayServeException:
        return AanaJSONResponse(
            content={
                "error": {"message": f"The model `{request.model}` does not exist."}
            },
            status_code=404,
        )

    # Check if the deployment is a chat model
    if not hasattr(handle, "chat") or not hasattr(handle, "chat_stream"):
        return AanaJSONResponse(
            content={
                "error": {"message": f"The model `{request.model}` does not exist."}
            },
            status_code=404,
        )

    dialog = ChatDialog(
        messages=request.messages,
    )

    sampling_params = SamplingParams(
        temperature=request.temperature,
        max_tokens=request.max_tokens,
        top_p=request.top_p,
    )

    if request.stream:
        return StreamingResponse(
            _async_chat_completions(handle, dialog, sampling_params),
            media_type="application/x-ndjson",
        )
    else:
        response = await handle.chat(dialog=dialog, sampling_params=sampling_params)
        return {
            "id": f"chatcmpl-{uuid4().hex}",
            "object": "chat.completion",
            "model": request.model,
            "created": int(time.time()),
            "choices": [{"index": 0, "message": response["message"]}],
        }



from typing import Annotated

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema


class ApiKey(BaseModel):
    """Pydantic model for API key entity.

    Attributes:
        api_key (str): The API key.
        user_id (str): ID of the user who owns this API key.
        subscription_id (str): ID of the associated subscription.
        is_subscription_active (bool): Whether the subscription is active (credits are available).
        is_admin (bool): Whether the user is an admin.
        hmac_secret (str | None): The secret key for HMAC signature generation.
    """

    api_key: str
    user_id: str
    subscription_id: str
    is_subscription_active: bool
    is_admin: bool
    hmac_secret: str | None


ApiKeyType = SkipJsonSchema[Annotated[ApiKey, Field(default=None)]]
"""
Type with optional API key information.

Can be None if API service is disabled. Otherwise, it will be an instance of `ApiKey`.

Attributes:
    api_key (str): The API key.
    user_id (str): The user ID.
    subscription_id (str): The subscription ID.
    is_subscription_active (bool): Flag indicating if the subscription is active.
"""



import logging
from typing import Annotated

from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, Field

from aana.api.security import UserIdDependency
from aana.configs.settings import settings as aana_settings
from aana.core.models.task import TaskInfo
from aana.storage.models.task import Status as TaskStatus
from aana.storage.repository.task import TaskRepository
from aana.storage.session import GetDbDependency

logger = logging.getLogger(__name__)

router = APIRouter(tags=["tasks"], include_in_schema=aana_settings.task_queue.enabled)

# Response models


class TaskList(BaseModel):
    """Response for a list of tasks."""

    tasks: list[TaskInfo] = Field(..., description="The list of tasks.")


# fmt: off
class TaskCount(BaseModel):
    """Response for a count of tasks by status."""

    created: int | None = Field(None, description="The number of tasks in the CREATED status.")
    assigned: int | None = Field(None, description="The number of tasks in the ASSIGNED status.")
    completed: int | None = Field(None, description="The number of tasks in the COMPLETED status.")
    running: int | None = Field(None, description="The number of tasks in the RUNNING status.")
    failed: int | None = Field(None, description="The number of tasks in the FAILED status.")
    not_finished: int | None = Field(None, description="The number of tasks in the NOT_FINISHED status.")
    total: int = Field(..., description="The total number of tasks.")
# fmt: on

# Endpoints


@router.get(
    "/tasks/count",
    summary="Count Tasks",
    description="Count tasks per status.",
    response_model_exclude_none=True,
)
async def count_tasks(db: GetDbDependency, user_id: UserIdDependency) -> TaskCount:
    """Count tasks by status."""
    task_repo = TaskRepository(db)
    counts = task_repo.count(user_id=user_id)
    return TaskCount(**counts)


@router.get(
    "/tasks/{task_id}",
    summary="Get Task Status",
    description="Get the task status by task ID.",
)
async def get_task(
    task_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> TaskInfo:
    """Get the task with the given ID."""
    task_repo = TaskRepository(db)
    task = task_repo.read(task_id, check=False)
    if not task or task.user_id != user_id:
        raise HTTPException(
            status_code=404,
            detail="Task not found",
        )
    return TaskInfo.from_entity(task)


@router.get(
    "/tasks",
    summary="List Tasks",
    description="List all tasks.",
)
async def list_tasks(
    db: GetDbDependency,
    user_id: UserIdDependency,
    status: Annotated[
        TaskStatus | None,
        Field(description="Filter tasks by status. If None, all tasks are returned."),
    ] = None,
    page: Annotated[int, Field(description="The page number.")] = 1,
    per_page: Annotated[int, Field(description="The number of tasks per page.")] = 100,
) -> TaskList:
    """List all tasks."""
    task_repo = TaskRepository(db)
    tasks = task_repo.get_tasks(
        user_id=user_id, status=status, limit=per_page, offset=(page - 1) * per_page
    )
    return TaskList(tasks=[TaskInfo.from_entity(task) for task in tasks])


@router.delete(
    "/tasks/{task_id}",
    summary="Delete Task",
    description="Delete the task by task ID.",
)
async def delete_task(
    task_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> TaskInfo:
    """Delete the task with the given ID."""
    task_repo = TaskRepository(db)
    task = task_repo.read(task_id, check=False)
    if not task or task.user_id != user_id:
        raise HTTPException(
            status_code=404,
            detail="Task not found",
        )
    if task.status in (TaskStatus.RUNNING, TaskStatus.ASSIGNED):
        raise HTTPException(
            status_code=400,
            detail="Cannot delete a running or assigned task.",
        )
    task = task_repo.delete(task_id)
    if task is None:
        raise HTTPException(
            status_code=404,
            detail="Task not found",
        )
    return TaskInfo.from_entity(task)


@router.post(
    "/tasks/{task_id}/retry",
    summary="Retry Failed Task",
    description="Retry a failed task by resetting its status to CREATED.",
)
async def retry_task(
    task_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> TaskInfo:
    """Retry a failed task by resetting its status."""
    task_repo = TaskRepository(db)
    task = task_repo.read(task_id, check=False)
    if not task or task.user_id != user_id:
        raise HTTPException(
            status_code=404,
            detail="Task not found",
        )
    if task.status != TaskStatus.FAILED:
        raise HTTPException(
            status_code=400,
            detail="Only failed tasks can be retried",
        )

    updated_task = task_repo.retry_task(task.id)
    return TaskInfo.from_entity(updated_task)


# Legacy endpoints (to be removed in the future)


@router.get(
    "/tasks/get/{task_id}",
    summary="Get Task Status (Legacy)",
    description="Get the task status by task ID (Legacy endpoint).",
    deprecated=True,
)
async def get_task_legacy(
    task_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> TaskInfo:
    """Get the task with the given ID (Legacy endpoint)."""
    task_repo = TaskRepository(db)
    task = task_repo.read(task_id)
    if not task or task.user_id != user_id:
        raise HTTPException(status_code=404, detail="Task not found")
    return TaskInfo.from_entity(task)


class TaskId(BaseModel):
    """Task ID (Legacy).

    Attributes:
        id (str): The task ID.
    """

    task_id: str = Field(..., description="The task ID.")


@router.get(
    "/tasks/delete/{task_id}",
    summary="Delete Task (Legacy)",
    description="Delete the task by task ID (Legacy endpoint).",
    deprecated=True,
)
async def delete_task_legacy(
    task_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> TaskId:
    """Delete the task with the given ID (Legacy endpoint)."""
    task_repo = TaskRepository(db)
    task = task_repo.read(task_id)
    if not task or task.user_id != user_id:
        raise HTTPException(status_code=404, detail="Task not found")
    task = task_repo.delete(task_id)
    return TaskId(task_id=str(task.id))



from sqlalchemy import Boolean
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from aana.core.models.api_service import ApiKey
from aana.storage.models.base import TimeStampEntity, timestamp


class ApiServiceBase(DeclarativeBase):
    """Base class."""

    pass


class ApiKeyEntity(ApiServiceBase, TimeStampEntity):
    """Table for API keys."""

    __tablename__ = "api_keys"
    __entity_name__ = "api key"

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True)
    key_id: Mapped[str] = mapped_column(
        nullable=False, unique=True, comment="The API key id in api gateway"
    )
    user_id: Mapped[str] = mapped_column(
        nullable=False, index=True, comment="ID of the user who owns this API key"
    )
    api_key: Mapped[str] = mapped_column(
        nullable=False, index=True, unique=True, comment="The API key"
    )
    is_admin: Mapped[bool] = mapped_column(
        Boolean, nullable=False, default=False, comment="Whether the user is an admin"
    )
    subscription_id: Mapped[str] = mapped_column(
        nullable=False, comment="ID of the associated subscription"
    )
    is_subscription_active: Mapped[bool] = mapped_column(
        Boolean,
        nullable=False,
        default=True,
        comment="Whether the subscription is active (credits are available)",
    )
    hmac_secret: Mapped[str] = mapped_column(
        nullable=True, comment="The secret key for HMAC signature generation"
    )
    expired_at: Mapped[timestamp] = mapped_column(
        nullable=False, comment="The expiration date of the API key"
    )

    def __repr__(self) -> str:
        """String representation of the API key."""
        return (
            f"<APIKeyEntity(id={self.id}, "
            f"key_id={self.key_id}, "
            f"user_id={self.user_id}, "
            f"api_key={self.api_key}, "
            f"is_admin={self.is_admin}, "
            f"subscription_id={self.subscription_id}, "
            f"hmac_secret={self.hmac_secret}, "
            f"is_subscription_active={self.is_subscription_active}), "
            f"expired_at={self.expired_at}, "
            f"created_at={self.created_at}, "
            f"updated_at={self.updated_at}>"
        )

    def to_model(self) -> ApiKey:
        """Convert the object to a dictionary."""
        return ApiKey(
            api_key=self.api_key,
            user_id=self.user_id,
            is_admin=self.is_admin,
            subscription_id=self.subscription_id,
            is_subscription_active=self.is_subscription_active,
            hmac_secret=self.hmac_secret,
        )



from aana.exceptions.core import BaseException


class ApiKeyNotProvided(BaseException):
    """Exception raised when the API key is not provided."""

    def __init__(self):
        """Initialize the exception."""
        self.message = "API key not provided"
        super().__init__(message=self.message)

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, ())


class ApiKeyNotFound(BaseException):
    """Exception raised when the API key is not found.

    Attributes:
        key (str): the API key that was not found
    """

    def __init__(self, key: str):
        """Initialize the exception.

        Args:
            key (str): the API key that was not found
        """
        self.key = key
        self.message = f"API key {key} not found"
        super().__init__(key=key, message=self.message)

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.key,))


class InactiveSubscription(BaseException):
    """Exception raised when the subscription is inactive (e.g. credits are not available).

    Attributes:
        key (str): the API key with inactive subscription
    """

    def __init__(self, key: str):
        """Initialize the exception.

        Args:
            key (str): the API key with inactive subscription
        """
        self.key = key
        self.message = (
            f"API key {key} has an inactive subscription. Check your credits."
        )
        super().__init__(key=key, message=self.message)

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.key,))


class AdminOnlyAccess(BaseException):
    """Exception raised when the user does not have enough permissions."""

    def __init__(self):
        """Initialize the exception."""
        self.message = "Admin only access"
        super().__init__(message=self.message)

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, ())


class ApiKeyValidationFailed(BaseException):
    """Exception raised when the API key validation fails."""

    def __init__(self):
        """Initialize the exception."""
        self.message = "API key validation failed"
        super().__init__(message=self.message)

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, ())



import asyncio
import hashlib
import hmac
import json
import logging
from typing import Any

import httpx
from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, ConfigDict, Field, HttpUrl
from tenacity import retry, stop_after_attempt, wait_exponential

from aana.api.security import UserIdDependency
from aana.configs.settings import settings as aana_settings
from aana.storage.models.api_key import ApiKeyEntity
from aana.storage.models.task import TaskEntity
from aana.storage.models.webhook import WebhookEntity, WebhookEventType
from aana.storage.repository.webhook import WebhookRepository
from aana.storage.session import GetDbDependency, get_session

logger = logging.getLogger(__name__)

router = APIRouter(tags=["webhooks"])


# Request models


class WebhookRegistrationRequest(BaseModel):
    """Request to register a webhook."""

    url: HttpUrl = Field(
        ..., description="The URL to which the webhook will send requests."
    )
    events: list[WebhookEventType] = Field(
        None,
        description="The events to subscribe to. If None, the webhook is subscribed to all events.",
    )

    model_config = ConfigDict(extra="forbid")


class WebhookUpdateRequest(BaseModel):
    """Request to update a webhook."""

    url: HttpUrl | None = Field(None, description="New URL for the webhook.")
    events: list[WebhookEventType] | None = Field(
        None, description="New list of events to subscribe to."
    )

    model_config = ConfigDict(extra="forbid")


# Response models


class WebhookRegistrationResponse(BaseModel):
    """Response for a webhook registration."""

    id: str | None
    message: str


class WebhookResponse(BaseModel):
    """Response for a webhook registration."""

    id: str = Field(..., description="The webhook ID.")
    url: HttpUrl = Field(
        ..., description="The URL to which the webhook will send requests."
    )
    events: list[WebhookEventType] = Field(
        ..., description="The events that the webhook is subscribed to."
    )

    @classmethod
    def from_entity(cls, webhook: WebhookEntity) -> "WebhookResponse":
        """Create a WebhookResponse from a WebhookEntity."""
        return WebhookResponse(
            id=str(webhook.id),
            url=str(webhook.url),
            events=webhook.events,
        )


class WebhookListResponse(BaseModel):
    """Response for a list of webhooks."""

    webhooks: list[WebhookResponse] = Field(..., description="The list of webhooks.")


# Webhook Models


class TaskStatusChangeWebhookPayload(BaseModel):
    """Payload for a task status change webhook."""

    task_id: str
    status: str
    result: Any | None
    num_retries: int


class WebhookBody(BaseModel):
    """Body for a task status change webhook."""

    event: WebhookEventType
    payload: TaskStatusChangeWebhookPayload


# Webhook functions


def generate_hmac_signature(body: dict, user_id: str | None) -> str:
    """Generate HMAC signature for a payload for a given user.

    Args:
        body (dict): The webhook body.
        user_id (str | None): The user ID associated with the payload.

    Returns:
        str: The generated HMAC signature.
    """
    # Use the default secret if no user ID is provided
    secret = aana_settings.webhook.hmac_secret
    # Get the user-specific secret if a user ID is provided
    if user_id:
        with get_session() as session:
            api_key_info = (
                session.query(ApiKeyEntity).filter_by(user_id=user_id).first()
            )
            if api_key_info and api_key_info.hmac_secret:
                secret = api_key_info.hmac_secret

    payload_str = json.dumps(body, separators=(",", ":"))
    return hmac.new(secret.encode(), payload_str.encode(), hashlib.sha256).hexdigest()


@retry(
    stop=stop_after_attempt(aana_settings.webhook.retry_attempts),
    wait=wait_exponential(),
    reraise=True,
)
async def send_webhook_request(url: str, body: dict, headers: dict):
    """Send a webhook request with retries.

    Args:
        url (str): The webhook URL.
        body (dict): The body of the request.
        headers (dict): The headers to include in the request.
    """
    async with httpx.AsyncClient(timeout=1.0) as client:
        response = await client.post(url, json=body, headers=headers)
        response.raise_for_status()


async def send_webhook_request_with_retry(url: str, body: dict, headers: dict):
    """Send a webhook request with retries.

    Args:
        url (str): The webhook URL.
        body (dict): The body of the request.
        headers (dict): The headers to include in the request.
    """
    try:
        await send_webhook_request(url, body, headers)
    except Exception:
        logger.exception(f"Failed to send webhook request to {url}.")


async def trigger_webhooks(
    event: WebhookEventType, body: WebhookBody, user_id: str | None
):
    """Trigger webhooks for an event.

    Args:
        event (WebhookEventType): The event type.
        body (WebhookBody): The body of the webhook request.
        user_id (str | None): The user ID associated with the event.
    """
    with get_session() as session:
        webhook_repo = WebhookRepository(session)
        webhooks = webhook_repo.get_webhooks(user_id, event)
        body_dict = body.model_dump()

        for webhook in webhooks:
            signature = generate_hmac_signature(body_dict, user_id)
            headers = {"X-Signature": signature}
            asyncio.create_task(  # noqa: RUF006
                send_webhook_request_with_retry(webhook.url, body_dict, headers)
            )


async def trigger_task_webhooks(event: WebhookEventType, task: TaskEntity):
    """Trigger webhooks for a task event.

    Args:
        event (WebhookEventType): The event type.
        task (TaskEntity): The task entity.
    """
    payload = TaskStatusChangeWebhookPayload(
        task_id=str(task.id),
        status=task.status,
        result=task.result,
        num_retries=task.num_retries,
    )
    body = WebhookBody(event=event, payload=payload)
    await trigger_webhooks(event, body, task.user_id)


# Webhook endpoints


@router.post("/webhooks", status_code=201)
async def create_webhook(
    request: WebhookRegistrationRequest,
    db: GetDbDependency,
    user_id: UserIdDependency,
) -> WebhookResponse:
    """This endpoint is used to register a webhook."""
    webhook_repo = WebhookRepository(db)
    webhook = WebhookEntity(
        user_id=user_id,
        url=str(request.url),
        events=request.events,
    )
    webhook = webhook_repo.save(webhook)
    return WebhookResponse.from_entity(webhook)


@router.get("/webhooks")
async def list_webhooks(
    db: GetDbDependency, user_id: UserIdDependency
) -> WebhookListResponse:
    """This endpoint is used to list all registered webhooks."""
    webhook_repo = WebhookRepository(db)
    webhooks = webhook_repo.get_webhooks(user_id, None)
    return WebhookListResponse(
        webhooks=[WebhookResponse.from_entity(webhook) for webhook in webhooks]
    )


@router.get("/webhooks/{webhook_id}")
async def get_webhook(
    webhook_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> WebhookResponse:
    """This endpoint is used to fetch a webhook by ID."""
    webhook_repo = WebhookRepository(db)
    webhook = webhook_repo.read(webhook_id, check=False)
    if not webhook or webhook.user_id != user_id:
        raise HTTPException(status_code=404, detail="Webhook not found")
    return WebhookResponse.from_entity(webhook)


@router.put("/webhooks/{webhook_id}")
async def update_webhook(
    webhook_id: str,
    request: WebhookUpdateRequest,
    db: GetDbDependency,
    user_id: UserIdDependency,
) -> WebhookResponse:
    """This endpoint is used to update a webhook."""
    webhook_repo = WebhookRepository(db)
    webhook = webhook_repo.read(webhook_id, check=False)
    if not webhook or webhook.user_id != user_id:
        raise HTTPException(status_code=404, detail="Webhook not found")
    if request.url is not None:
        webhook.url = str(request.url)
    if request.events is not None:
        webhook.events = request.events
    webhook = webhook_repo.save(webhook)
    return WebhookResponse.from_entity(webhook)


@router.delete("/webhooks/{webhook_id}")
async def delete_webhook(
    webhook_id: str, db: GetDbDependency, user_id: UserIdDependency
) -> WebhookResponse:
    """This endpoint is used to delete a webhook."""
    webhook_repo = WebhookRepository(db)
    webhook = webhook_repo.read(webhook_id, check=False)
    if not webhook or webhook.user_id != user_id:
        raise HTTPException(status_code=404, detail="Webhook not found")
    webhook = webhook_repo.delete(webhook.id)
    return WebhookResponse.from_entity(webhook)



# Adapted from: https://github.com/deepset-ai/haystack
#
#  SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0

from types import TracebackType

from lazy_imports.try_import import _DeferredImportExceptionContextManager

DEFAULT_IMPORT_ERROR_MSG = "Try 'pip install {}'"


class LazyImport(_DeferredImportExceptionContextManager):
    """Wrapper on top of lazy_import's _DeferredImportExceptionContextManager.

    It adds the possibility to customize the error messages.
    """

    def __init__(self, message: str = DEFAULT_IMPORT_ERROR_MSG) -> None:
        """Initialize the context manager."""
        super().__init__()
        self.import_error_msg = message

    def __exit__(
        self,
        exc_type: type[Exception] | None = None,
        exc_value: Exception | None = None,
        traceback: TracebackType | None = None,
    ) -> bool | None:
        """Exit the context manager.

        Args:
            exc_type:
                Raised exception type. :obj:`None` if nothing is raised.
            exc_value:
                Raised exception object. :obj:`None` if nothing is raised.
            traceback:
                Associated traceback. :obj:`None` if nothing is raised.

        Returns:
            :obj:`None` if nothing is deferred, otherwise :obj:`True`.
            :obj:`True` will suppress any exceptions avoiding them from propagating.

        """
        if isinstance(exc_value, ImportError):
            message = (
                f"Failed to import '{exc_value.name}'. {self.import_error_msg.format(exc_value.name)}. "
                f"Original error: {exc_value}"
            )
            self._deferred = (exc_value, message)
            return True
        return None



import uuid
from enum import Enum

from sqlalchemy import JSON, UUID
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column

from aana.storage.models.base import BaseEntity, TimeStampEntity


class WebhookEventType(str, Enum):
    """Enum for webhook event types."""

    TASK_COMPLETED = "task.completed"
    TASK_FAILED = "task.failed"
    TASK_STARTED = "task.started"


class WebhookEntity(BaseEntity, TimeStampEntity):
    """Table for webhook items."""

    __tablename__ = "webhooks"

    id: Mapped[uuid.UUID] = mapped_column(
        UUID, primary_key=True, default=uuid.uuid4, comment="Webhook ID"
    )
    user_id: Mapped[str | None] = mapped_column(
        nullable=True, index=True, comment="The user ID associated with the webhook"
    )
    url: Mapped[str] = mapped_column(
        nullable=False, comment="The URL to which the webhook will send requests"
    )
    events: Mapped[list[str]] = mapped_column(
        JSON().with_variant(JSONB, "postgresql"),
        nullable=False,
        comment="List of events the webhook is subscribed to. If the list is empty, the webhook is subscribed to all events.",
    )

    def __repr__(self) -> str:
        """String representation of the webhook."""
        return (
            f"<WebhookEntity(id={self.id}, user_id={self.user_id}, "
            f"webhook_url={self.url}, events={self.events}, "
            f"created_at={self.created_at}, updated_at={self.updated_at})>"
        )



# ruff: noqa: S101, S113
import io
import json
from typing import TypedDict

import requests
from pydantic import BaseModel, ConfigDict, Field

from aana.api.api_generation import Endpoint
from aana.exceptions.runtime import UploadedFileNotFound


class FileUploadModel(BaseModel):
    """Model for a file upload input."""

    content: str | None = Field(
        None,
        description="The name of the file to upload.",
    )
    _file: bytes | None = None

    def set_files(self, files: dict[str, bytes]):
        """Set files."""
        if self.content:
            if self.content not in files:
                raise UploadedFileNotFound(self.content)
            self._file = files[self.content]

    model_config = ConfigDict(extra="forbid")


class FileUploadEndpointOutput(TypedDict):
    """The output of the file upload endpoint."""

    text: str


class FileUploadEndpoint(Endpoint):
    """File upload endpoint."""

    async def run(self, file: FileUploadModel) -> FileUploadEndpointOutput:
        """Upload a file.

        Args:
            file (FileUploadModel): The file to upload

        Returns:
            FileUploadEndpointOutput: The uploaded file
        """
        file = file._file
        return {"text": file.decode()}


deployments = []

endpoints = [
    {
        "name": "file_upload",
        "path": "/file_upload",
        "summary": "Upload a file",
        "endpoint_cls": FileUploadEndpoint,
    }
]


def test_file_upload_app(create_app):
    """Test the app with a file upload endpoint."""
    aana_app = create_app(deployments, endpoints)

    port = aana_app.port
    route_prefix = ""

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}{route_prefix}/api/ready")
    assert response.status_code == 200
    assert response.json() == {"ready": True}

    # Test lowercase endpoint
    # data = {"content": "file.txt"}
    data = {"file": {"content": "file.txt"}}
    file = b"Hello world! This is a test."
    files = {"file.txt": io.BytesIO(file)}
    response = requests.post(
        f"http://localhost:{port}{route_prefix}/file_upload",
        data={"body": json.dumps(data)},
        files=files,
    )
    assert response.status_code == 200, response.text
    text = response.json().get("text")
    assert text == file.decode()



from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError

from aana.api.exception_handler import (
    aana_exception_handler,
    validation_exception_handler,
)
from aana.configs.settings import settings as aana_settings
from aana.exceptions.api_service import (
    ApiKeyNotFound,
    ApiKeyNotProvided,
    ApiKeyValidationFailed,
)
from aana.storage.models.api_key import ApiKeyEntity
from aana.storage.session import get_session

app = FastAPI()

app.add_exception_handler(ValidationError, validation_exception_handler)
app.add_exception_handler(RequestValidationError, validation_exception_handler)
app.add_exception_handler(Exception, aana_exception_handler)


@app.middleware("http")
async def api_key_check(request: Request, call_next):
    """Middleware to check the API key and subscription status."""
    excluded_paths = ["/openapi.json", "/docs", "/redoc"]
    if request.url.path in excluded_paths:
        return await call_next(request)

    if aana_settings.api_service.enabled:
        api_key = request.headers.get("x-api-key")

        if not api_key:
            raise ApiKeyNotProvided()

        with get_session() as session:
            try:
                api_key_info = (
                    session.query(ApiKeyEntity).filter_by(api_key=api_key).first()
                )
            except Exception as e:
                raise ApiKeyValidationFailed() from e

            if not api_key_info:
                raise ApiKeyNotFound(key=api_key)

            request.state.api_key_info = api_key_info.to_model()

    response = await call_next(request)
    return response



"""Added user_id to tasks.

Revision ID: d40eba8ebc4c
Revises: b9860676dd49
Create Date: 2025-01-23 14:17:04.394863

"""
from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d40eba8ebc4c"
down_revision: str | None = "b9860676dd49"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade database to this revision from previous."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "user_id",
                sa.String(),
                nullable=True,
                comment="ID of the user who launched the task",
            )
        )
        batch_op.create_index(batch_op.f("ix_tasks_status"), ["status"], unique=False)
        batch_op.create_index(batch_op.f("ix_tasks_user_id"), ["user_id"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database from this revision to previous."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_tasks_user_id"))
        batch_op.drop_index(batch_op.f("ix_tasks_status"))
        batch_op.drop_column("user_id")

    # ### end Alembic commands ###



from typing import Any
from uuid import UUID

import orjson
from fastapi import APIRouter
from fastapi.openapi.utils import get_openapi
from ray import serve

from aana.api.api_generation import Endpoint, add_custom_schemas_to_openapi_schema
from aana.api.app import app
from aana.api.event_handlers.event_manager import EventManager
from aana.api.exception_handler import custom_exception_handler
from aana.api.responses import AanaJSONResponse
from aana.api.security import AdminAccessDependency
from aana.core.models.api import DeploymentStatus, SDKStatus, SDKStatusResponse
from aana.routers.openai import router as openai_router
from aana.routers.task import router as task_router
from aana.routers.webhook import (
    WebhookEventType,
    trigger_task_webhooks,
)
from aana.routers.webhook import router as webhook_router
from aana.storage.models.task import Status as TaskStatus
from aana.storage.repository.task import TaskRepository
from aana.storage.session import get_session


@serve.deployment(ray_actor_options={"num_cpus": 0.1})
@serve.ingress(app)
class RequestHandler:
    """This class is used to handle requests to the Aana application."""

    ready = False

    def __init__(
        self,
        app_name: str,
        endpoints: list[Endpoint],
        deployments: list[str],
        routers: list[APIRouter] | None = None,
    ):
        """Constructor.

        Args:
            app_name (str): The name of the application.
            endpoints (dict): List of endpoints for the request.
            deployments (list[str]): List of deployment names for the app.
            routers (list[APIRouter]): List of FastAPI routers to include in the app.
        """
        self.app_name = app_name
        self.endpoints = endpoints
        self.deployments = deployments

        # Include the default routers
        app.include_router(webhook_router)  # For webhook management
        app.include_router(task_router)  # For task management
        app.include_router(openai_router)  # For OpenAI-compatible API

        # Include the custom routers (from Aana Apps)
        if routers is not None:
            for router in routers:
                app.include_router(router)

        self.event_manager = EventManager()
        self.custom_schemas: dict[str, dict] = {}
        for endpoint in self.endpoints:
            endpoint.register(
                app=app,
                custom_schemas=self.custom_schemas,
                event_manager=self.event_manager,
            )

        app.openapi = self.custom_openapi
        self.ready = True
        self.running_tasks = set()

    def custom_openapi(self) -> dict[str, Any]:
        """Returns OpenAPI schema, generating it if necessary."""
        if app.openapi_schema:
            return app.openapi_schema
        # TODO: populate title and version from package info
        openapi_schema = get_openapi(title="Aana", version="0.1.0", routes=app.routes)
        openapi_schema = add_custom_schemas_to_openapi_schema(
            openapi_schema=openapi_schema, custom_schemas=self.custom_schemas
        )
        app.openapi_schema = openapi_schema
        return app.openapi_schema

    async def execute_task(self, task_id: str | UUID) -> Any:
        """Execute a task.

        Args:
            task_id (str | UUID): The ID of the task.

        Returns:
            Any: The response from the endpoint.
        """
        try:
            self.running_tasks.add(task_id)
            with get_session() as session:
                task_repo = TaskRepository(session)
                task = task_repo.read(task_id)
                path = task.endpoint
                kwargs = task.data

                task = task_repo.update_status(task_id, TaskStatus.RUNNING, 0)
                await trigger_task_webhooks(WebhookEventType.TASK_STARTED, task)

            for e in self.endpoints:
                if e.path == path:
                    endpoint = e
                    break
            else:
                raise ValueError(f"Endpoint {path} not found")  # noqa: TRY003, TRY301

            if not endpoint.initialized:
                await endpoint.initialize()

            if endpoint.is_streaming_response():
                out = [item async for item in endpoint.run(**kwargs)]
            else:
                out = await endpoint.run(**kwargs)

            with get_session() as session:
                task = TaskRepository(session).update_status(
                    task_id, TaskStatus.COMPLETED, 100, out
                )
                await trigger_task_webhooks(WebhookEventType.TASK_COMPLETED, task)
        except Exception as e:
            error_response = custom_exception_handler(None, e)
            error = orjson.loads(error_response.body)
            with get_session() as session:
                task = TaskRepository(session).update_status(
                    task_id, TaskStatus.FAILED, 0, error
                )
                await trigger_task_webhooks(WebhookEventType.TASK_FAILED, task)
        else:
            return out
        finally:
            self.running_tasks.remove(task_id)

    @app.get("/api/ready", tags=["system"])
    async def is_ready(self):
        """The endpoint for checking if the application is ready.

        Real reason for this endpoint is to make automatic endpoint generation work.
        If RequestHandler doesn't have any endpoints defined manually,
        then the automatic endpoint generation doesn't work.
        #TODO: Find a better solution for this.

        Returns:
            AanaJSONResponse: The response containing the ready status.
        """
        return AanaJSONResponse(content={"ready": self.ready})

    async def check_health(self):
        """Check the health of the application."""
        # Heartbeat for the running tasks
        with get_session() as session:
            task_repo = TaskRepository(session)
            task_repo.heartbeat(self.running_tasks)

    @app.get("/api/status", response_model=SDKStatusResponse, tags=["system"])
    async def status(self, is_admin: AdminAccessDependency) -> SDKStatusResponse:
        """The endpoint for checking the status of the application."""
        app_names = [
            self.app_name,
            *self.deployments,
        ]  # the list of Ray Serve apps that belong to this Aana app
        serve_status = serve.status()
        app_statuses = {
            app_name: app_status
            for app_name, app_status in serve_status.applications.items()
            if app_name in app_names
        }

        app_status_message = ""
        if any(
            app.status in ["DEPLOY_FAILED", "UNHEALTHY", "NOT_STARTED"]
            for app in app_statuses.values()
        ):
            sdk_status = SDKStatus.UNHEALTHY
            error_messages = []
            for app_name, app_status in app_statuses.items():
                if app_status.status in ["DEPLOY_FAILED", "UNHEALTHY"]:
                    for (
                        deployment_name,
                        deployment_status,
                    ) in app_status.deployments.items():
                        error_messages.append(
                            f"Error: {deployment_name} ({app_name}): {deployment_status.message}"
                        )
            app_status_message = "\n".join(error_messages)
        elif all(app.status == "RUNNING" for app in app_statuses.values()):
            sdk_status = SDKStatus.RUNNING
        elif any(
            app.status in ["DEPLOYING", "DELETING"] for app in app_statuses.values()
        ):
            sdk_status = SDKStatus.DEPLOYING
        else:
            sdk_status = SDKStatus.UNHEALTHY
            app_status_message = "Unknown status"

        deployment_statuses = {}
        for app_name, app_status in app_statuses.items():
            messages = []
            for deployment_name, deployment_status in app_status.deployments.items():
                if deployment_status.message:
                    messages.append(
                        f"{deployment_name} ({app_name}): {deployment_status.message}"
                    )
            message = "\n".join(messages)

            deployment_statuses[app_name] = DeploymentStatus(
                status=app_status.status, message=message
            )

        return SDKStatusResponse(
            status=sdk_status,
            message=app_status_message,
            deployments=deployment_statuses,
        )



import asyncio
import os
from collections.abc import AsyncGenerator
from enum import Enum
from pathlib import Path
from typing import Any, cast

import nvidia.cudnn.lib
import torch
from pydantic import BaseModel, ConfigDict, Field
from ray import serve
from typing_extensions import TypedDict

from aana.core.models.asr import (
    AsrSegment,
    AsrTranscription,
    AsrTranscriptionInfo,
)
from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_protected_fields
from aana.core.models.vad import VadSegment
from aana.core.models.whisper import BatchedWhisperParams, WhisperParams
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.utils.lazy_import import LazyImport

with LazyImport(
    "Run 'pip install faster-whisper' or 'pip install aana[asr]'"
) as faster_whisper_import:
    from faster_whisper import BatchedInferencePipeline, WhisperModel

# Workaround for CUDNN issue with cTranslate2:
cudnn_path = str(Path(nvidia.cudnn.lib.__file__).parent)
os.environ["LD_LIBRARY_PATH"] = (
    cudnn_path + "/:" + os.environ.get("LD_LIBRARY_PATH", "")
)


class WhisperComputeType(str, Enum):
    """The data type used by whisper models.

    See [cTranslate2 docs on quantization](https://opennmt.net/CTranslate2/quantization.html#quantize-on-model-conversion)
    for more information.

    Available types:
        - INT8
        - INT8_FLOAT32
        - INT8_FLOAT16
        - INT8_BFLOAT16
        - INT16
        - FLOAT16
        - BFLOAT16
        - FLOAT32
    """

    INT8 = "int8"
    INT8_FLOAT32 = "int8_float32"
    INT8_FLOAT16 = "int8_float16"
    INT8_BFLOAT16 = "int8_bfloat16"
    INT16 = "int16"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    FLOAT32 = "float32"


class WhisperModelSize(str, Enum):
    """The whisper model.

    Available models:
        - TINY
        - TINY_EN
        - BASE
        - BASE_EN
        - SMALL
        - SMALL_EN
        - MEDIUM
        - MEDIUM_EN
        - LARGE_V1
        - LARGE_V2
        - LARGE_V3
        - LARGE
        - TURBO
    """

    TINY = "tiny"
    TINY_EN = "tiny.en"
    BASE = "base"
    BASE_EN = "base.en"
    SMALL = "small"
    SMALL_EN = "small.en"
    MEDIUM = "medium"
    MEDIUM_EN = "medium.en"
    LARGE_V1 = "large-v1"
    LARGE_V2 = "large-v2"
    LARGE_V3 = "large-v3"
    LARGE = "large"
    TURBO = "mobiuslabsgmbh/faster-whisper-large-v3-turbo"  # SYSTRAN/faster-whisper do not have turbo model in HF.


class WhisperConfig(BaseModel):
    """The configuration for the whisper deployment from faster-whisper.

    Attributes:
        model_size (WhisperModelSize | str): The whisper model size. Defaults to WhisperModelSize.TURBO.
        compute_type (WhisperComputeType): The compute type. Defaults to WhisperComputeType.FLOAT16.
    """

    model_size: WhisperModelSize | str = Field(
        default=WhisperModelSize.TURBO,
        description="The whisper model size. Either one of WhisperModelSize or HuggingFace model ID in Ctranslate2 format.",
    )
    compute_type: WhisperComputeType = Field(
        default=WhisperComputeType.FLOAT16, description="The compute type."
    )

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


class WhisperOutput(TypedDict):
    """The output of the whisper model.

    Attributes:
        segments (list[AsrSegment]): The ASR segments.
        transcription_info (AsrTranscriptionInfo): The ASR transcription info.
        transcription (AsrTranscription): The ASR transcription.
    """

    segments: list[AsrSegment]
    transcription_info: AsrTranscriptionInfo
    transcription: AsrTranscription


class WhisperBatchOutput(TypedDict):
    """The output of the whisper model for a batch of inputs.

    Attributes:
        segments (list[list[AsrSegment]]): The ASR segments for each audio.
        transcription_info (list[AsrTranscriptionInfo]): The ASR transcription info for each audio.
        transcription (list[AsrTranscription]): The ASR transcription for each audio.
    """

    segments: list[list[AsrSegment]]
    transcription_info: list[AsrTranscriptionInfo]
    transcription: list[AsrTranscription]


@serve.deployment(max_ongoing_requests=1)
class WhisperDeployment(BaseDeployment):
    """Deployment to serve Whisper models from faster-whisper."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        The configuration should conform to the WhisperConfig schema.
        """
        faster_whisper_import.check()
        config_obj = WhisperConfig(**config)
        self.model_size = config_obj.model_size
        self.model_name = "whisper_" + self.model_size
        self.compute_type = config_obj.compute_type
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = WhisperModel(
            self.model_size, device=self.device, compute_type=self.compute_type
        )
        self.batched_model = BatchedInferencePipeline(
            model=self.model,
        )

    async def check_health(self):
        """Check the health of the deployment and clear torch cache to prevent memory leaks."""
        torch.cuda.empty_cache()
        await super().check_health()

    async def transcribe(
        self, audio: Audio, params: WhisperParams | None = None
    ) -> WhisperOutput:
        """Transcribe the audio with the whisper model.

        Args:
            audio (Audio): The audio to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Returns:
            WhisperOutput: The transcription output as a dictionary:

                - segments (List[AsrSegment]): The ASR segments.
                - transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                - transcription (AsrTranscription): The ASR transcription.

        Raises:
            InferenceException: If the inference fails.
        """
        asr_segments = []
        asr_transcription_info = None
        async for output in self.transcribe_stream(audio, params):
            asr_segments.extend(output["segments"])
            asr_transcription_info = output["transcription_info"]

        transcription = "".join([seg.text for seg in asr_segments])
        asr_transcription = AsrTranscription(text=transcription)

        return WhisperOutput(
            segments=asr_segments,
            transcription_info=asr_transcription_info,
            transcription=asr_transcription,
        )

    @exception_handler
    async def transcribe_stream(
        self, audio: Audio, params: WhisperParams | None = None
    ) -> AsyncGenerator[WhisperOutput, None]:
        """Transcribe the audio with the whisper model in a streaming fashion.

        Args:
            audio (Audio): The audio to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Yields:
            WhisperOutput: The transcription output as a dictionary:

                - segments (list[AsrSegment]): The ASR segments.
                - transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                - transcription (AsrTranscription): The ASR transcription.
        """
        if not params:
            params = WhisperParams()
        audio_array = audio.get_numpy()
        if not audio_array.any():
            # For silent audios/no audio tracks, return empty output with language as silence
            yield WhisperOutput(
                segments=[],
                transcription_info=AsrTranscriptionInfo(
                    language="silence", language_confidence=1.0
                ),
                transcription=AsrTranscription(text=""),
            )
        else:
            try:
                segments, info = self.model.transcribe(
                    audio_array, **params.model_dump()
                )
                asr_transcription_info = AsrTranscriptionInfo.from_whisper(info)
                for segment in segments:
                    await asyncio.sleep(0)
                    asr_segments = [AsrSegment.from_whisper(segment)]
                    asr_transcription = AsrTranscription(text=segment.text)

                    yield WhisperOutput(
                        segments=asr_segments,
                        transcription_info=asr_transcription_info,
                        transcription=asr_transcription,
                    )
            except Exception as e:
                raise InferenceException(self.model_name, str(e)) from e

    async def transcribe_batch(
        self, audio_batch: list[Audio], params: WhisperParams | None = None
    ) -> WhisperBatchOutput:
        """Transcribe the batch of audios with the Whisper model.

        Args:
            audio_batch (list[Audio]): The batch of audios to transcribe.
            params (WhisperParams): The parameters for the whisper model.

        Returns:
            WhisperBatchOutput: The transcription output as a dictionary:

                - segments (list[list[AsrSegment]]): The ASR segments for each audio.
                - transcription_info (list[AsrTranscriptionInfo]): The ASR transcription info for each audio.
                - transcription (list[AsrTranscription]): The ASR transcription for each audio.

        Raises:
            InferenceException: If the inference fails.
        """
        if not params:
            params = WhisperParams()
        segments: list[list[AsrSegment]] = []
        infos: list[AsrTranscriptionInfo] = []
        transcriptions: list[AsrTranscription] = []
        for audio in audio_batch:
            output = await self.transcribe(audio, params)
            segments.append(cast(list[AsrSegment], output["segments"]))
            infos.append(cast(AsrTranscriptionInfo, output["transcription_info"]))
            transcriptions.append(cast(AsrTranscription, output["transcription"]))

        return WhisperBatchOutput(
            segments=segments, transcription_info=infos, transcription=transcriptions
        )

    @exception_handler
    async def transcribe_in_chunks(
        self,
        audio: Audio,
        vad_segments: list[VadSegment] | None = None,
        batch_size: int = 4,
        params: BatchedWhisperParams | None = None,
    ) -> AsyncGenerator[WhisperOutput, None]:
        """Transcribe a single audio by segmenting it into chunks (4x faster) in streaming mode.

        Args:
            audio (Audio): The audio to transcribe.
            vad_segments (list[VadSegment]| None): List of vad segments to guide batching of the audio data.
            batch_size (int): Maximum batch size for the batched inference.
            params (BatchedWhisperParams | None): The parameters for the batched inference with whisper model.

        Yields:
            WhisperOutput: The transcription output as a dictionary:
                segments (list[AsrSegment]): The ASR segments.
                transcription_info (AsrTranscriptionInfo): The ASR transcription info.
                transcription (AsrTranscription): The ASR transcription.

        Raises:
            InferenceException: If the inference fails.
        """
        if not params:
            params = BatchedWhisperParams()
        audio_array = audio.get_numpy()
        if vad_segments:
            sampling_rate = self.model.feature_extractor.sampling_rate
            vad_input = [seg.to_whisper_dict() for seg in vad_segments]
            vad_input = [
                {
                    "start": int(seg["start"] * sampling_rate),
                    "end": int(seg["end"] * sampling_rate),
                }
                for seg in vad_input
            ]

        if not audio_array.any():
            # For silent audios/no audio tracks, return empty output with language as silence
            yield WhisperOutput(
                segments=[],
                transcription_info=AsrTranscriptionInfo(
                    language="silence", language_confidence=1.0
                ),
                transcription=AsrTranscription(text=""),
            )
        else:
            try:
                segments, info = self.batched_model.transcribe(
                    audio_array,
                    clip_timestamps=vad_input if vad_segments else None,
                    batch_size=batch_size,
                    **params.model_dump(),
                )
                asr_transcription_info = AsrTranscriptionInfo.from_whisper(info)
                for segment in segments:
                    await asyncio.sleep(0)
                    asr_segments = [AsrSegment.from_whisper(segment)]
                    asr_transcription = AsrTranscription(text=segment.text)
                    yield WhisperOutput(
                        segments=asr_segments,
                        transcription_info=asr_transcription_info,
                        transcription=asr_transcription,
                    )
            except Exception as e:
                raise InferenceException(self.model_name, str(e)) from e



# ruff: noqa: F401
# We need to import all db models here and, other than in the class definitions
# themselves, only import them from aana.models.db directly. The reason for
# this is the way SQLAlchemy's declarative base works. You can use forward
# references like `parent = reference("Parent", backreferences="child")`, but the
# forward reference needs to have been resolved before the first constructor
# is called so that SqlAlchemy "knows" about it.
# See:
# https://docs.pylonsproject.org/projects/pyramid_cookbook/en/latest/database/sqlalchemy.html#importing-all-sqlalchemy-models
# (even if not using Pyramid)

from aana.storage.models.base import BaseEntity
from aana.storage.models.caption import CaptionEntity
from aana.storage.models.media import MediaEntity
from aana.storage.models.task import TaskEntity
from aana.storage.models.transcript import TranscriptEntity
from aana.storage.models.video import VideoEntity
from aana.storage.models.webhook import WebhookEntity

__all__ = [
    "BaseEntity",
    "CaptionEntity",
    "MediaEntity",
    "TranscriptEntity",
    "VideoEntity",
]



import inspect
import logging
import time
import uuid
from collections.abc import AsyncGenerator, Callable
from dataclasses import dataclass
from enum import Enum
from inspect import isasyncgenfunction
from typing import Annotated, Any, get_origin

from fastapi import FastAPI, Form, Query, Request
from fastapi.responses import StreamingResponse
from pydantic import ConfigDict, Field, ValidationError, create_model
from pydantic.main import BaseModel
from starlette.datastructures import UploadFile as StarletteUploadFile

from aana.api.event_handlers.event_handler import EventHandler
from aana.api.event_handlers.event_manager import EventManager
from aana.api.exception_handler import custom_exception_handler
from aana.api.responses import AanaJSONResponse
from aana.api.security import require_active_subscription, require_admin_access
from aana.configs.settings import settings as aana_settings
from aana.core.models.api_service import ApiKey
from aana.core.models.exception import ExceptionResponseModel
from aana.storage.repository.task import TaskRepository
from aana.storage.session import get_session

logger = logging.getLogger(__name__)


def get_default_values(func):
    """Get the default values for the function arguments."""
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


class DeferOption(str, Enum):
    """Enum for defer option.

    Attributes:
        ALWAYS (str): Always defer. Endpoints with this option will always be defer execution to the task queue.
        NEVER (str): Never defer. Endpoints with this option will never be defer execution to the task queue.
        OPTIONAL (str): Optionally defer. Endpoints with this option can be defer execution to the task queue if
            the defer query parameter is set to True.
    """

    ALWAYS = "always"
    NEVER = "never"
    OPTIONAL = "optional"


@dataclass
class Endpoint:
    """Class used to represent an endpoint.

    Attributes:
        name (str): Name of the endpoint.
        path (str): Path of the endpoint (e.g. "/video/transcribe").
        summary (str): Description of the endpoint that will be shown in the API documentation.
        admin_required (bool): Flag indicating if the endpoint requires admin access.
        active_subscription_required (bool): Flag indicating if the endpoint requires an active subscription.
        defer_option (DeferOption): Defer option for the endpoint (always, never, optional).
        event_handlers (list[EventHandler] | None): The list of event handlers to register for the endpoint.
    """

    name: str
    path: str
    summary: str
    admin_required: bool = False
    active_subscription_required: bool = False
    defer_option: DeferOption = DeferOption.OPTIONAL
    initialized: bool = False
    event_handlers: list[EventHandler] | None = None

    async def initialize(self):
        """Initialize the endpoint.

        Redefine this method to add initialization logic for the endpoint (e.g. create a handle to the deployment).
        Call super().initialize() to ensure the endpoint is initialized.

        Example:
            ```python
            async def initialize(self):
                await super().initialize()
                self.asr_handle = await AanaDeploymentHandle.create("whisper_deployment")
            ```
        """
        self.initialized = True

    async def run(self, *args, **kwargs):
        """The main method of the endpoint that is called when the endpoint receives a request.

        Redefine this method to implement the logic of the endpoint.

        Example:
            ```python
            async def run(self, video: VideoInput) -> WhisperOutput:
                video_obj: Video = await run_remote(download_video)(video_input=video)
                transcription = await self.asr_handle.transcribe(audio=audio)
                return transcription
            ```
        """
        raise NotImplementedError

    def register(
        self, app: FastAPI, custom_schemas: dict[str, dict], event_manager: EventManager
    ):
        """Register the endpoint in the FastAPI application.

        Args:
            app (FastAPI): FastAPI application.
            custom_schemas (dict[str, dict]): Dictionary containing custom schemas.
            event_manager (EventManager): Event manager for the application.
        """
        RequestModel = self.get_request_model()
        ResponseModel = self.get_response_model()

        if self.event_handlers:
            for handler in self.event_handlers:
                event_manager.register_handler_for_events(handler, [self.path])

        route_func = self.__create_endpoint_func(
            RequestModel=RequestModel,
            event_manager=event_manager,
        )

        app.post(
            self.path,
            name=self.name,
            summary=self.summary,
            operation_id=self.name,
            response_model=ResponseModel,
            responses={
                400: {"model": ExceptionResponseModel},
            },
        )(route_func)
        custom_schemas[self.name] = RequestModel.model_json_schema()

    def __generate_model_name(self, suffix: str) -> str:
        """Generate a Pydantic model name based on a given suffix.

        Args:
            suffix (str): Suffix for the model name (e.g. "Request", "Response").

        Returns:
            str: Generated model name.
        """
        return self.__class__.__name__ + suffix

    def __arg_to_field(self, arg_name: str, arg_type: Any) -> tuple[Any, Any]:
        """Convert an argument to a Pydantic field.

        Args:
            arg_name (str): Name of the argument.
            arg_type (Any): Type of the argument.

        Returns:
            Tuple[Any, Any]: Tuple containing the Pydantic model and the Pydantic field.
        """
        # if data model is None or Any, set it to Any
        if arg_type is None or arg_type == Any:
            return (Any, Field(None))

        if get_origin(arg_type) is Annotated:
            default = get_default_values(self.run).get(arg_name)
            if default is not None:
                return (arg_type, default)
            else:
                return (arg_type, ...)

        # try to instantiate the data model
        # to see if any of the fields are required
        try:
            data_model_instance = arg_type()
        except TypeError as e:
            raise ValueError(  # noqa: TRY003
                f"Invalid type for argument {arg_name}: {arg_type}. "
                "Consider using Annotated[..., Field(...)] or Pydantic models."
            ) from e
        except ValidationError:
            # if we can't instantiate the data model
            # it means that it has required fields
            return (arg_type, ...)
        else:
            return (arg_type, data_model_instance)

    def __get_input_fields(self) -> dict[str, tuple[Any, Any]]:
        """Generate fields for the request Pydantic model based on the function annotations.

        Returns:
            dict[str, tuple[Any, Field]]: Dictionary of fields for the request Pydantic model.
        """
        fields = {}
        for arg_name, arg_type in self.run.__annotations__.items():
            if arg_name == "return":
                continue
            fields[arg_name] = self.__arg_to_field(arg_name, arg_type)
        return fields

    def get_request_model(self) -> type[BaseModel]:
        """Generate the request Pydantic model for the endpoint.

        Returns:
            type[BaseModel]: Request Pydantic model.
        """
        model_name = self.__generate_model_name("Request")
        input_fields = self.__get_input_fields()
        return create_model(model_name, **input_fields)

    def __get_output_fields(self) -> dict[str, tuple[Any, Any]]:
        """Generate fields for the response Pydantic model based on the function annotations.

        Returns:
            dict[str, tuple[Any, Any]]: Dictionary of fields for the response Pydantic model.
        """
        try:
            if self.is_streaming_response():
                return_type = self.run.__annotations__["return"].__args__[0]
            else:
                return_type = self.run.__annotations__["return"]
        except (AttributeError, KeyError) as e:
            raise ValueError("Endpoint function must have a return annotation.") from e  # noqa: TRY003
        fields = {}
        for arg_name, arg_type in return_type.__annotations__.items():
            fields[arg_name] = (arg_type, Field(None))
        return fields

    def get_response_model(self) -> type[BaseModel]:
        """Generate the response Pydantic model for the endpoint.

        Returns:
            type[BaseModel]: Response Pydantic model.
        """
        model_name = self.__generate_model_name("Response")
        output_fields = self.__get_output_fields()
        return create_model(
            model_name, **output_fields, __config__=ConfigDict(extra="forbid")
        )

    @classmethod
    def is_streaming_response(cls) -> bool:
        """Check if the endpoint returns a streaming response.

        Returns:
            bool: True if the endpoint returns a streaming response, False otherwise.
        """
        return isasyncgenfunction(cls.run)

    def __create_endpoint_func(  # noqa: C901
        self,
        RequestModel: type[BaseModel],
        event_manager: EventManager | None = None,
    ) -> Callable:
        """Create a function for routing an endpoint."""
        # Copy path to a bound variable so we don't retain an external reference
        bound_path = self.path

        async def route_func_body(  # noqa: C901
            request: Request,
            body: str,
            files: dict[str, bytes] | None = None,
            defer=False,
        ):
            if not self.initialized:
                await self.initialize()

            if event_manager:
                event_manager.handle(bound_path, defer=defer)

            # parse form data as a pydantic model and validate it
            data = RequestModel.model_validate_json(body)

            # Add api_key_info to the body if API service is enabled
            api_key_info: ApiKey | None = None
            if aana_settings.api_service.enabled:
                api_key_info = request.state.api_key_info
                api_key_field = next(
                    (
                        field_name
                        for field_name, field in RequestModel.model_fields.items()
                        if field.annotation is ApiKey
                    ),
                    None,
                )
                if api_key_field:
                    setattr(data, api_key_field, api_key_info)

            # if the input requires file upload, add the files to the data
            if files:
                for field_name in data.model_fields:
                    field_value = getattr(data, field_name)
                    if hasattr(field_value, "set_files"):
                        field_value.set_files(files)

            # We have to do this instead of data.dict() because
            # data.dict() will convert all nested models to dicts
            # and we want to keep them as pydantic models
            data_dict = {}
            for field_name in data.model_fields:
                field_value = getattr(data, field_name)
                data_dict[field_name] = field_value

            if defer:
                if not aana_settings.task_queue.enabled:
                    raise RuntimeError("Task queue is not enabled.")  # noqa: TRY003

                with get_session() as session:
                    task = TaskRepository(session).save(
                        endpoint=bound_path,
                        data=data_dict,
                        user_id=api_key_info.user_id if api_key_info else None,
                    )
                    return AanaJSONResponse(content={"task_id": str(task.id)})

            if isasyncgenfunction(self.run):

                async def generator_wrapper() -> AsyncGenerator[bytes, None]:
                    """Serializes the output of the generator using ORJSONResponseCustom."""
                    try:
                        async for output in self.run(**data_dict):
                            yield AanaJSONResponse(content=output).body
                    except Exception as e:
                        yield custom_exception_handler(None, e).body

                return StreamingResponse(
                    generator_wrapper(), media_type="application/json"
                )
            else:
                try:
                    output = await self.run(**data_dict)
                except Exception as e:
                    return custom_exception_handler(None, e)
                return AanaJSONResponse(content=output)

        async def route_func(
            request: Request,
            body: str = Form(...),
            defer: bool = Query(
                description="Defer execution of the endpoint to the task queue.",
                default=False,
                include_in_schema=aana_settings.task_queue.enabled
                and self.defer_option == DeferOption.OPTIONAL,
            ),
        ):
            if aana_settings.api_service.enabled and self.admin_required:
                require_admin_access(request)

            if aana_settings.api_service.enabled and self.active_subscription_required:
                require_active_subscription(request)

            if self.defer_option == DeferOption.ALWAYS:
                defer = True
            elif self.defer_option == DeferOption.NEVER:
                defer = False

            form_data = await request.form()

            # Parse files from the form data
            files: dict[str, bytes] = {}
            for field_name, field_value in form_data.items():
                if isinstance(field_value, StarletteUploadFile):
                    files[field_name] = await field_value.read()

            return await route_func_body(
                request=request, body=body, files=files, defer=defer
            )

        return route_func

    def send_usage_event(
        self, api_key: ApiKey, metric: str, properties: dict[str, Any]
    ):
        """Send a usage event to the LAGO API service.

        Args:
            api_key (ApiKey): The API key information.
            metric (str): The metric code.
            properties (dict): The properties of the event (usage data, e.g. {"count": 10}).
        """
        from lago_python_client.client import Client
        from lago_python_client.models import Event
        from tenacity import (
            retry,
            retry_if_exception_type,
            stop_after_attempt,
        )

        @retry(
            stop=stop_after_attempt(3),
            retry=retry_if_exception_type(Exception),
        )
        def send_event_with_retry(client, event):
            return client.events.create(event)

        if not aana_settings.api_service.enabled:
            logger.warning("API service is not enabled. Skipping usage event.")
            return

        if (
            not aana_settings.api_service.lago_url
            or not aana_settings.api_service.lago_api_key
        ):
            logger.warning(
                "LAGO API service URL or API key is not set. Skipping usage event."
            )
            return

        try:
            client = Client(
                api_key=aana_settings.api_service.lago_api_key,
                api_url=aana_settings.api_service.lago_url,
            )

            event = Event(
                transaction_id=str(uuid.uuid4()),
                code=metric,
                external_subscription_id=api_key.subscription_id,
                timestamp=time.time(),
                properties=properties,
            )

            send_event_with_retry(client, event)
        except Exception as e:
            logger.error(
                f"Failed to send usage event after retries: {e}", exc_info=True
            )


def add_custom_schemas_to_openapi_schema(
    openapi_schema: dict[str, Any], custom_schemas: dict[str, Any]
) -> dict[str, Any]:
    """Add custom schemas to the openapi schema.

    File upload is that FastAPI doesn't support Pydantic models in multipart requests.
    There is a discussion about it on FastAPI discussion forum.
    See https://github.com/tiangolo/fastapi/discussions/8406
    The topic starter suggests a workaround.
    The workaround is to use Forms instead of Pydantic models in the endpoint definition and
    then convert the Forms to Pydantic models in the endpoint itself
    using parse_raw_as function from Pydantic.
    Since Pydantic model isn't used in the endpoint definition,
    the API documentation will not be generated automatically.
    So the workaround also suggests updating the API documentation manually
    by overriding the openapi method of a FastAPI application.

    Args:
        openapi_schema (dict): The openapi schema.
        custom_schemas (dict): The custom schemas.

    Returns:
        dict: The openapi schema with the custom schemas added.
    """
    if "$defs" not in openapi_schema:
        openapi_schema["$defs"] = {}
    for schema_name, schema in custom_schemas.items():
        # if we have a definitions then we need to move them out to the top level of the schema
        if "$defs" in schema:
            openapi_schema["$defs"].update(schema["$defs"])
            del schema["$defs"]
        openapi_schema["components"]["schemas"][f"Body_{schema_name}"]["properties"][
            "body"
        ] = schema
    return openapi_schema
